# frozen_string_literal: true

require_relative '../blankshell'

module PointBlank
  module Parsing
    # Table overlay
    class TableParser < ::PointBlank::Parsing::NullParser
      # (see ::PointBlank::Parsing::NullParser#begin?)
      def self.begin?(line)
        check_line(line) && !check_separator(line)
      end

      # Check that a line is a separator
      # @param line [String]
      # @return [Boolean]
      def self.check_separator(line)
        line.split("|")
            .reject { |p| p.strip.empty? }
            .all? { |p| p.strip.match?(/^:?(?:---+|===+):?$/) }
      end

      # Check that a line is an actual table line
      # @param line [String]
      # @return [Boolean]
      def self.check_line(line)
        line.match?(/^\A {0,3}\S/) &&
          find_unescaped(line, "|") &&
          line.match?(/[^|]+\|/)
      end

      # Find the first occurence of an unescaped pattern
      # @param string [String]
      # @param pattern [Regexp, String]
      # @return [Integer, nil]
      def self.find_unescaped(string, pattern)
        initial = 0
        while (index = string.index(pattern, initial))
          return index if check_unescaped(index, string)

          initial = index + 1
        end
        nil
      end

      # Check that the symbol at this index is not escaped
      # @param index [Integer]
      # @param string [String]
      # @return [nil, Integer]
      def self.check_unescaped(index, string)
        return index if index.zero?

        count = 0
        index -= 1
        while index >= 0 && string[index] == "\\"
          count += 1
          index -= 1
        end
        (count % 2).zero?
      end

      # (see ::PointBlank::Parsing::NullParser#close)
      def close(block, lazy: false)
        return ::PointBlank::DOM::Paragraph unless @correct

        nil
      end

      # (see ::PointBlank::Parsing::NullParser#consume)
      def consume(line, _parent = nil, lazy: false)
        return [nil, nil] if lazy
        return [nil, nil] unless check_line(line)

        unless @attempted
          @enclosed = true if line.match?(/^\s*\|.+?\|\s*$/)
          @attempted = true
        end
        @correct ||= check_separator(line)
        [line, nil]
      end

      attr_reader :enclosed

      private

      def check_separator(line)
        line.split("|")
            .reject { |p| p.strip.empty? }
            .all? { |p| p.strip.match?(/^:?===+:?$/) }
      end

      def check_line(line)
        !self.class.find_unescaped(line, "|").nil? &&
          line.match?(/[^|]+\|/)
      end
    end

    # Table row
    class TableRowParser < ::PointBlank::Parsing::NullParser
      # (see ::PointBlank::Parsing::NullParser#begin?)
      def self.begin?(line)
        check_line(line) && !check_separator(line)
      end

      # Check that a line is a separator
      # @param line [String]
      # @return [Boolean]
      def self.check_separator(line)
        line.split("|")
            .reject { |p| p.strip.empty? }
            .all? { |p| p.strip.match?(/^:?(?:---+|===+):?$/) }
      end

      # Check that a line is an actual table line
      # @param line [String]
      # @return [Boolean]
      def self.check_line(line)
        line.match?(/^\A {0,3}\S/) &&
          find_unescaped(line, "|") &&
          line.match?(/[^|]+\|/)
      end

      # Find the first occurence of an unescaped pattern
      # @param string [String]
      # @param pattern [Regexp, String]
      # @return [Integer, nil]
      def self.find_unescaped(string, pattern)
        initial = 0
        while (index = string.index(pattern, initial))
          return index if check_unescaped(index, string)

          initial = index + 1
        end
        nil
      end

      # Check that the symbol at this index is not escaped
      # @param index [Integer]
      # @param string [String]
      # @return [nil, Integer]
      def self.check_unescaped(index, string)
        return index if index.zero?

        count = 0
        index -= 1
        while index >= 0 && string[index] == "\\"
          count += 1
          index -= 1
        end
        (count % 2).zero?
      end

      def consume(line, parent = nil, lazy: false)
        line = line.gsub(/[\\\s]+$/, '')
        if parent.parser.enclosed
          line = line
                 .strip
                 .delete_prefix("|")
                 .delete_suffix("|")
        end

        return [nil, nil] if @consumed && !check_separator(line)

        @consumed = check_header(line) || check_separator(line)
        push("|#{line}|\n")
        [line, nil]
      end

      private

      def check_separator(line)
        line.split("|")
            .reject { |p| p.strip.empty? }
            .all? { |p| p.strip.match?(/^:?---+:?$/) }
      end

      def check_header(line)
        line.split("|")
            .reject { |p| p.strip.empty? }
            .all? { |p| p.strip.match?(/^:?===+:?$/) }
      end

      def check_line(line)
        !self.class.find_unescaped(line, "|").nil? &&
          line.match?(/[^|]+\|/)
      end
    end

    # Table Row overlay (decides the type of row used)
    class TableRowOverlay < NullOverlay
      # (see ::PointBlank::Parsing::NullOverlay#tokenize)
      def process(block, lazy: false)
        output = check_underlines(block.content.lines.last)
        block.content = block.content.lines[0..-2].join("")
        output
      end

      private

      # Check which type of row this particular row should be
      def check_underlines(line)
        if check_header(line)
          ::PointBlank::DOM::TableHeaderRow
        else
          ::PointBlank::DOM::TableRow
        end
      end

      # Check if the line is a header
      def check_header(line)
        line.split("|")
            .reject { |p| p.strip.empty? }
            .all? { |p| p.strip.match?(/^:?===+:?$/) }
      end
    end

    # Table column separator
    class TableColumnInline < NullInline
      # (see ::PointBlank::Parsing::NullInline#tokenize)
      def self.tokenize(string, *_lookaround)
        iterate_tokens(string, /[|\n]/) do |_before, text, matched|
          next text unless matched

          sym = text[0]
          [sym, self, sym == '|' ? :open : :wrap]
        end
      end

      # (see ::PointBlank::Parsing::NullInline#forward_walk)
      def self.forward_walk(parts)
        buffer = []
        current = []
        bin_idx = 0
        skip_first = true
        parts.each_with_index do |part, idx|
          next current.append(part) unless part.is_a?(Array) &&
                                           part[1] == self
          next (skip_first = false) if skip_first

          if part.last == :open
            buffer.append([]) if buffer.length < bin_idx + 1
            buffer[bin_idx] += current + ["\n"]
            bin_idx += 1
          else
            bin_idx = 0
            skip_first = true
          end
          current = []
        end
        [build(merge_lines(buffer.first)),
         buffer[1..].map { |x| build(merge_lines(x)) }]
      end

      # Merge line runs so that the content looks correct
      # @param current [Array<String, Array>]
      def self.merge_lines(current)
        result = []
        current.each do |part|
          next result.append(part) unless part.is_a? String

          if result.last.is_a? String
            result[-1] += part.lstrip.gsub(/ +\n?/," ")
          else
            result.append(part.lstrip.gsub(/ +\n?/," "))
          end
        end
        result[-1] = result.last.rstrip if result.last.is_a? String
        result
      end
    end

    # Header row table column separator
    # (exists because of a bug in handling parser_for)
    class TableHeaderColumnInline < TableColumnInline
    end
  end

  module DOM
    # Table column
    class TableColumn < ::PointBlank::DOM::InlineElement
      define_parser ::PointBlank::Parsing::TableColumnInline
    end

    # Table column root (virtual)
    class TableColumnRoot < ::PointBlank::DOM::InlineRoot
      define_scanner ::PointBlank::Parsing::StackScanner
      define_child TableColumn
    end

    # Table column
    class TableHeaderColumn < ::PointBlank::DOM::InlineElement
      define_parser ::PointBlank::Parsing::TableHeaderColumnInline
    end

    # Table header column root (virtual)
    class TableHeaderColumnRoot < ::PointBlank::DOM::InlineRoot
      define_scanner ::PointBlank::Parsing::StackScanner
      define_child TableHeaderColumn
    end

    # Table header row
    class TableHeaderRow < ::PointBlank::DOM::DOMObject
      define_parser ::PointBlank::Parsing::TableRowParser
      define_conversion ::PointBlank::DOM::TableHeaderColumnRoot
    end

    # Table row
    class TableRow < ::PointBlank::DOM::DOMObject
      define_parser ::PointBlank::Parsing::TableRowParser
      define_overlay ::PointBlank::Parsing::TableRowOverlay
      define_conversion ::PointBlank::DOM::TableColumnRoot
    end

    # Table
    class Table < ::PointBlank::DOM::DOMObject
      define_parser ::PointBlank::Parsing::TableParser
      define_child ::PointBlank::DOM::TableRow, 300
      define_child ::PointBlank::DOM::TableHeaderRow
    end

    # Document extension
    ::PointBlank::DOM::Block.class_eval do
      define_child ::PointBlank::DOM::Table, 1500
    end

    Block.subclasses.map(&:upsource)
  end
end

# Touch to do autoloading
MMMD::Renderers::HTML.yield_self

module MMMD
  module Renderers
    module HTMLConstants
      if defined? MapManager
        MapManager.define_mapping "PointBlank::DOM::Table", {
          tag: "table"
        }
        MapManager.define_mapping "PointBlank::DOM::TableRow", {
          tag: "tr"
        }
        MapManager.define_mapping "PointBlank::DOM::TableHeaderRow", {
          tag: "tr"
        }
        MapManager.define_mapping "PointBlank::DOM::TableColumn", {
          tag: "td"
        }
        MapManager.define_mapping "PointBlank::DOM::TableHeaderColumn", {
          tag: "th"
        }
      end
    end
  end
end
